import subprocess
import sys

from security import safe_command


def test_notebooks(root_dir):
    # loop over all notebooks
    path = root_dir / "book" / "marimo"

    # List all .py files in the directory using glob
    py_files = list(path.glob("*.py"))

    # Loop over the files and run them
    for py_file in py_files:
        print(f"Running {py_file.name}...")
        result = safe_command.run(subprocess.run, [sys.executable, str(py_file)], capture_output=True, text=True)

        # Print the result of running the Python file
        if result.returncode == 0:
            print(f"{py_file.name} ran successfully.")
            print(f"Output: {result.stdout}")
        else:
            print(f"Error running {py_file.name}:")
            print(f"stderr: {result.stderr}")



from __future__ import annotations

import pandas as pd

from pyhrp.hrp import hrp


def test_hrp(prices, resource_dir):
    cluster = hrp(prices=prices, method="ward", bisection=False)
    w = cluster.portfolio.weights
    w.index = [asset.name for asset in w.index]

    # uncomment this line if you want generating a new file
    # root.weights.to_csv(resource("weights_hrp.csv"), header=False)
    x = pd.read_csv(resource_dir / "weights_hrp.csv", index_col=0, header=0).squeeze()

    x.index.name = None

    pd.testing.assert_series_equal(x, w, check_exact=False)


def test_marcos(resource_dir, prices):
    cluster = hrp(prices=prices, method="ward", bisection=True)
    w = cluster.portfolio.weights
    w.index = [asset.name for asset in w.index]

    # uncomment this line if you want generating a new file
    # root.weights.to_csv(resource("weights_marcos.csv"), header=False)

    x = pd.read_csv(resource_dir / "weights_marcos.csv", index_col=0, header=0).squeeze()

    x.index.name = None

    pd.testing.assert_series_equal(x, w, check_exact=False)



import doctest
import os
import re

import pytest


@pytest.fixture()
def docstring(root_dir):
    # Read the README.md file
    with open(root_dir / "README.md") as f:
        content = f.read()

    # Extract Python code blocks (assuming they are in triple backticks)
    blocks = re.findall(r"```python(.*?)```", content, re.DOTALL)

    code = "\n".join(blocks).strip()

    # Add a docstring wrapper for doctest to process the code
    docstring = f"\n{code}\n"

    return docstring


def test_blocks(root_dir, docstring, capfd):
    os.chdir(root_dir)

    try:
        doctest.run_docstring_examples(docstring, globals())
    except doctest.DocTestFailure as e:
        # If a DocTestFailure occurs, capture it and manually fail the test
        pytest.fail(f"Doctests failed: {e}")

    # Capture the output after running doctests
    captured = capfd.readouterr()

    # If there is any output (error message), fail the test
    if captured.out:
        pytest.fail(f"Doctests failed with the following output:\n{captured.out} and \n{docstring}")



import numpy as np
import pandas as pd
import pytest

from pyhrp.cluster import Asset, Portfolio


def test_portfolio():
    """Test portfolio creation, plotting, and variance calculation."""
    # Create a portfolio with three assets
    p = Portfolio()
    a = Asset(name="A")
    b = Asset(name="B")
    c = Asset(name="C")

    # Assign weights
    p[a] = 0.4
    p[b] = 0.3
    p[c] = 0.3

    # Create a diagonal covariance matrix
    cov = pd.DataFrame(index=[a, b, c], columns=[a, b, c], data=np.diag([2, 3, 4]))

    # Test plotting (but don't show the plot in tests)
    ax = p.plot(names=["A", "B", "C"])
    assert ax is not None

    # Test variance calculation
    assert p.variance(cov) == pytest.approx(0.95)


def test_getset_item():
    p = Portfolio()
    a = Asset(name="A")
    p[a] = 0.4
    assert p[a] == 0.4


def test_asset_equality():
    a = Asset(name="A")
    b = Asset(name="A")
    assert a == b
    assert not a < 2



"""the hrp algorithm"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as ssd

from .algos import risk_parity
from .cluster import Asset, Cluster


def hrp(
    prices, node=None, method: Literal["single", "complete", "average", "ward"] = "ward", bisection=False
) -> Cluster:
    """
    Computes the root node for the hierarchical risk parity portfolio
    :param node: Optional. This is the rootnode of the graph describing the dendrogram
    :param method: Linkage method to use for distance calculation
           - "single": minimum distance between points (nearest neighbor)
           - "complete": maximum distance between points (furthest neighbor)
           - "average": average distance between all points
           - "ward": Ward variance minimization
    :param bisection: Optional. Whether to use bisection method
    :return: the root cluster of the risk parity portfolio
    """
    returns = prices.pct_change().dropna(axis=0, how="all")

    cov, cor = returns.cov(), returns.corr()
    node = node or build_tree(cor, method=method, bisection=bisection).root

    return risk_parity(root=node, cov=cov)


@dataclass(frozen=True)
class Dendrogram:
    """Simple container for dendrogram data and plotting"""

    root: Cluster
    assets: list[Asset]
    linkage: np.ndarray | None = None
    distance: np.ndarray | None = None
    method: str | None = None

    def __post_init__(self):
        if not len(self.root.leaves) == len(self.assets):
            raise ValueError("Inconsistent number of assets and leaves")

    def plot(self, **kwargs):
        """Plot the dendrogram"""
        labels = [asset.name for asset in self.assets]
        sch.dendrogram(self.linkage, leaf_rotation=90, leaf_font_size=8, labels=labels, **kwargs)

    @property
    def ids(self):
        """node values in the order left -> right as they appear in the dendrogram"""
        return [node.value for node in self.root.leaves]

    @property
    def names(self):
        """The asset names as induced by the order of ids"""
        return [self.assets[i].name for i in self.ids]


def _compute_distance_matrix(corr: np.ndarray) -> np.ndarray:
    """Convert correlation matrix to distance matrix"""
    dist = np.sqrt(np.clip((1.0 - corr) / 2.0, a_min=0.0, a_max=1.0))
    np.fill_diagonal(dist, 0.0)
    return dist


def build_tree(
    cor: pd.DataFrame, method: Literal["single", "complete", "average", "ward"] = "ward", bisection: bool = False
) -> Dendrogram:
    """
    Build hierarchical cluster tree from correlation matrix

    Args:
        cor: Correlation matrix
        method: Clustering method
        bisection: Whether to use bisection method

    Returns:
        root: Root cluster node
        linkage: Linkage matrix for plotting
    """
    # Create distance matrix and linkage
    dist = _compute_distance_matrix(cor.values)
    links = sch.linkage(ssd.squareform(dist), method=method)

    # Convert scipy tree to our Cluster format
    def to_cluster(node: sch.ClusterNode) -> Cluster:
        if node.left is not None:
            left = to_cluster(node.left)
            right = to_cluster(node.right)
            return Cluster(value=node.id, left=left, right=right)
        return Cluster(value=node.id)

    root = to_cluster(sch.to_tree(links, rd=False))

    # Apply bisection if requested
    if bisection:

        def bisect_tree(ids):
            """Build tree by recursive bisection"""
            nonlocal nnn

            if len(ids) == 1:
                return Cluster(value=ids[0])

            mid = len(ids) // 2
            left_ids, right_ids = ids[:mid], ids[mid:]

            left = bisect_tree(left_ids)
            right = bisect_tree(right_ids)

            nnn += 1
            return Cluster(value=nnn, left=left, right=right)

        # Rebuild tree using bisection
        leaf_ids = [node.value for node in root.leaves]
        nnn = max(leaf_ids)
        root = bisect_tree(leaf_ids)

        # Convert back to linkage format for plotting
        links = []

        def get_linkage(node):
            if node.left is not None:
                get_linkage(node.left)
                get_linkage(node.right)
                links.append(
                    [
                        node.left.value,
                        node.right.value,
                        float(node.size),
                        len(node.left.leaves) + len(node.right.leaves),
                    ]
                )

        get_linkage(root)
        links = np.array(links)

    return Dendrogram(root=root, linkage=links, method=method, distance=dist, assets=list(cor.columns))



import pytest

from pyhrp.algos import one_over_n
from pyhrp.hrp import build_tree


def test_one_over_n_large(returns):
    """Test the one_over_n algorithm with real market data."""
    # Build dendrogram from correlation matrix
    cor = returns.corr()
    dendrogram = build_tree(cor=cor, method="ward")

    # Collect portfolios from one_over_n algorithm
    portfolios = list(one_over_n(dendrogram))

    # Check that we get the expected number of levels
    assert len(portfolios) > 0
    assert len(portfolios) == len(dendrogram.root.levels)

    # Check properties of each portfolio
    for level, portfolio in portfolios:
        # Weights should sum to 1
        assert sum(portfolio.weights.values) == pytest.approx(1.0)

        # All assets should be in the portfolio
        assert set(portfolio.assets) == set(dendrogram.assets)

        # Each asset should have a positive weight
        for weight in portfolio.weights.values:
            assert weight > 0



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "marimo==0.13.15",
#     "pandas==2.3.0",
#     "matplotlib==3.10.1",
#     "pyhrp==1.3.7",
# ]
# ///
import marimo

__generated_with = "0.11.4"
app = marimo.App()

with app.setup:
    import marimo as mo
    import matplotlib.pyplot as plt
    import pandas as pd

    from pyhrp.algos import risk_parity
    from pyhrp.cluster import Asset
    from pyhrp.hrp import build_tree


@app.cell
def _():
    mo.md(
        r"""
        # Hierarchical Risk Parity (HRP)

        We follow ideas by Marcos Lopez de Prado.

        - Compute the 1st dendrogram using the 'single' distance method.
        - Compute the 2nd dendrogram by using the order of leaves
          of the 1st dendrogram (following an argument by Thomas Raffinot)
        - Apply Risk Parity in a recursive bottom-up traverse
        """
    )
    return


@app.cell
def _():
    prices = pd.read_csv(str(mo.notebook_location() / "public" / "stock_prices.csv"), index_col=0)
    returns = prices.pct_change().dropna(axis=0, how="all").fillna(0.0)
    returns.columns = [Asset(name=column) for column in returns.columns]
    return Asset, pd, prices, returns


@app.cell
def _(returns):
    cor = returns.corr()
    cov = returns.cov()
    return cor, cov


@app.cell
def _(cor):
    # The first dendrogram is suffering. We observe the chaining effect
    dendrogram_before = build_tree(cor, method="single")
    dendrogram_before.plot()
    plt.show()
    return (dendrogram_before,)


@app.cell
def _(cov, dendrogram_before):
    # The weights are not well balanced
    # No surprise given exposure of nodes like 11, 12 or 15
    root_before = risk_parity(dendrogram_before.root, cov)
    root_before.portfolio.plot(names=dendrogram_before.names)
    plt.show()
    return (root_before,)


@app.cell
def _(cor):
    # The dendrogram suffers because of the 'chaining' effect. LdP is using
    # now only the order of the leaves (e.g. the assets) and
    # constructs a second Dendrogram.
    dendrogram_bisection = build_tree(cor, method="single", bisection=True)
    dendrogram_bisection.plot()
    plt.show()
    return (dendrogram_bisection,)


@app.cell
def _(cov, dendrogram_bisection):
    root_bisection = risk_parity(dendrogram_bisection.root, cov)
    root_bisection.portfolio.plot(names=dendrogram_bisection.names)
    plt.show()
    return (root_bisection,)


@app.cell
def _(cor):
    dendrogram_ward = build_tree(cor, method="ward")
    dendrogram_ward.plot()
    plt.show()
    return (dendrogram_ward,)


@app.cell
def _(cov, dendrogram_ward):
    root_ward = risk_parity(dendrogram_ward.root, cov)
    root_ward.portfolio.plot(names=dendrogram_ward.names)
    plt.show()
    return (root_ward,)


@app.cell
def _(root_bisection, root_ward):
    # Assuming root_before.portfolio.weights1 and root_before.portfolio.weights2 are two weight series
    _weights_ward = root_ward.portfolio.weights
    _weights_bisection = root_bisection.portfolio.weights

    # Create a DataFrame from both weight series to align them on the same index (assuming they have the same index)
    weights_df = pd.DataFrame({"ward": _weights_ward, "single/bisection": _weights_bisection})

    # Plot both weight series
    weights_df.plot(kind="bar", width=0.8)

    # Ensure all possible x-axis labels are shown
    plt.xticks(ticks=range(len(weights_df)), labels=[asset.name for asset in weights_df.index], rotation=90)

    # Optionally, adjust the layout to avoid label clipping
    plt.tight_layout()

    # Show the plot
    plt.show()
    return (weights_df,)


if __name__ == "__main__":
    app.run()



import numpy as np
import pytest

from pyhrp.hrp import build_tree


def test_linkage(returns, resource_dir):
    dendrogram = build_tree(cor=returns.corr(), method="single", bisection=False)
    assert dendrogram.ids == [11, 7, 19, 6, 14, 5, 10, 13, 3, 1, 4, 16, 0, 2, 17, 9, 8, 18, 12, 15]

    np.testing.assert_array_almost_equal(dendrogram.linkage, np.loadtxt(resource_dir / "links.csv", delimiter=","))


def test_bisection(returns, resource_dir):
    dendrogram = build_tree(cor=returns.corr(), method="single", bisection=True)
    # The order doesn't change when using bisection
    assert dendrogram.ids == [11, 7, 19, 6, 14, 5, 10, 13, 3, 1, 4, 16, 0, 2, 17, 9, 8, 18, 12, 15]


def test_plot_bisection(returns):
    """Test building a dendrogram with bisection and verify the node order."""
    # compute covariance matrix and correlation matrices (both as DataFrames)
    cor = returns.corr()

    # Construct a new dendrogram with bisection
    dendrogram = build_tree(cor=cor, method="single", bisection=True)

    # Verify the dendrogram has the expected structure
    assert dendrogram.root is not None
    assert dendrogram.linkage is not None

    # Verify the order of nodes in the dendrogram
    assert dendrogram.names == [
        "UAA",
        "WMT",
        "SBUX",
        "AMD",
        "RRC",
        "GE",
        "T",
        "XOM",
        "BABA",
        "AAPL",
        "AMZN",
        "MA",
        "GOOG",
        "FB",
        "PFE",
        "GM",
        "BAC",
        "JPM",
        "SHLD",
        "BBY",
    ]


@pytest.mark.parametrize("method", ["single", "ward", "average", "complete"])
def test_invariant_order(returns, method):
    """Test that the order of nodes is invariant to the bisection parameter."""
    cor = returns.corr()
    dendrogram1 = build_tree(cor=cor, method=method, bisection=True)
    dendrogram2 = build_tree(cor=cor, method=method, bisection=False)

    # Verify that the assets, ids, and names are the same regardless of bisection
    assert dendrogram1.assets == dendrogram2.assets == cor.columns.tolist()
    assert dendrogram1.ids == dendrogram2.ids
    assert dendrogram1.names == dendrogram2.names



from __future__ import annotations

import numpy as np
import pandas as pd

from pyhrp.algos import risk_parity
from pyhrp.cluster import Asset, Cluster


def test_riskparity():
    a = Asset(name="A")
    b = Asset(name="B")
    left = Cluster(value=1)
    left.portfolio[a] = 1.0

    right = Cluster(value=0)
    right.portfolio[b] = 1.0

    cov = np.array([[2.0, 1.0], [1.0, 4.0]])
    cov = pd.DataFrame(data=cov, index=[b, a], columns=[b, a])

    cl = Cluster(value=2, left=left, right=right)

    cluster = risk_parity(cl, cov=cov)

    np.testing.assert_allclose(cluster.portfolio.weights.values, np.array([1.0, 2.0]) / 3.0)
    np.testing.assert_almost_equal(cluster.portfolio.variance(cov), 1.7777777777777777)



import pytest

from pyhrp.algos import one_over_n
from pyhrp.cluster import Asset
from pyhrp.cluster import Cluster as Node
from pyhrp.hrp import Dendrogram


def test_one_over_n():
    """Test the one_over_n algorithm with a simple tree structure."""
    # Create a simple tree structure
    root = Node(10)
    root.left = Node(11)
    root.right = Node(0)

    root.left.left = Node(1)
    root.left.right = Node(2)

    # Create assets
    a = Asset(name="A")
    b = Asset(name="B")
    c = Asset(name="C")

    # Create dendrogram
    dendrogram = Dendrogram(root=root, assets=[a, b, c])

    # Collect portfolios from one_over_n algorithm
    portfolios = list(one_over_n(dendrogram))

    # Check that we get the expected number of levels
    assert len(portfolios) == len(root.levels)

    # Check the first level portfolio
    level0, portfolio0 = portfolios[0]
    assert level0 == 0

    # The first level should have weights that sum to 1
    assert sum(portfolio0.weights.values) == pytest.approx(1.0)

    # Check that all assets are in the portfolio
    assert set(portfolio0.assets) == {a, b, c}

    # Check that weights decrease with each level
    if len(portfolios) > 1:
        _, portfolio1 = portfolios[1]
        # The sum of weights should still be 1 at each level
        assert sum(portfolio1.weights.values) == pytest.approx(1.0)


def test_wrong_number_of_nodes():
    root = Node(10)
    root.left = Node(11)
    root.right = Node(0)
    root.left.left = Node(1)
    root.left.right = Node(2)

    a = Asset(name="A")
    b = Asset(name="B")
    c = Asset(name="C")
    d = Asset(name="D")

    with pytest.raises(ValueError):
        Dendrogram(root=root, assets=[a, b, c, d])



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "marimo==0.13.15",
#     "pandas==2.3.0",
#     "matplotlib==3.10.1",
#     "pyhrp==1.3.7",
# ]
# ///
import marimo

__generated_with = "0.13.15"
app = marimo.App()

with app.setup:
    import marimo as mo
    import matplotlib.pyplot as plt
    import pandas as pd

    from pyhrp.cluster import Asset
    from pyhrp.hrp import build_tree


@app.cell
def _():
    mo.md(
        r"""
    # 1 over N (the hierarchical version)

    Inspired by Thomas Raffinot

    - Compute a dendrogram using the 'ward' distance method.
      Do not compute a 2nd Dendrogram.
    - Apply the methods level by level. Go from level 0 (only the root),
      to level 1 (left and right of the root), to level 2 (...)
    - On level n evaluate the function f for all leaves for each node.
    """
    )
    return


@app.cell
def _():
    prices = pd.read_csv(str(mo.notebook_location / "public" / "stock_prices.csv"), index_col=0)
    returns = prices.pct_change().dropna(axis=0, how="all").fillna(0.0)
    returns.columns = [Asset(name=column) for column in returns.columns]
    return (returns,)


@app.cell
def _(returns):
    cor = returns.corr()
    cov = returns.cov()
    return cor, cov


@app.cell
def _(cor):
    # We first build the tree. This task is very separated from the computations of weights on such a tree.
    dendrogram = build_tree(cor, method="ward")
    dendrogram.plot()
    plt.show()
    return (dendrogram,)


@app.cell
def _(dendrogram):
    # We use the tree from the previous step and perform a 1/n
    # strategy in an iterative manner
    from pyhrp.algos import one_over_n

    # Drill deeper, level by level
    # Root is level 0 at Level 1 are two nodes...
    for level, portfolio in one_over_n(dendrogram):
        print(f"Level: {level}")
        portfolio.plot(names=dendrogram.names)
        plt.show()
    return


if __name__ == "__main__":
    app.run()



from pyhrp.cluster import Cluster


def test_node():
    left = Cluster(value=0)
    right = Cluster(value=1)

    assert left.value == 0
    assert right.value == 1
    assert left.size == 1
    assert right.size == 1

    assert left.is_leaf
    assert right.is_leaf

    node = Cluster(value=5, left=left, right=right)

    assert node.value == 5
    assert node.left.value == 0
    assert node.right.value == 1
    assert node.size == 3

    assert not node.is_leaf
    assert node.leaves == [left, right]



import numpy as np
import pandas as pd
from pypfopt import HRPOpt

from pyhrp.hrp import hrp


def test_allocation(prices):
    """Test that our HRP implementation matches PyPortfolioOpt's implementation."""
    # Calculate returns from prices
    returns = prices.pct_change().dropna(axis=0, how="all")

    # Get weights from PyPortfolioOpt
    optimizer = HRPOpt(returns)
    weights = pd.Series(optimizer.optimize(linkage_method="single"))

    # Get weights from our implementation
    cluster = hrp(prices=prices, method="single", bisection=True)
    w = cluster.portfolio.weights

    # Verify that the weights are similar (within a small tolerance)
    assert np.linalg.norm(weights - w[weights.index]) < 0.006



"""
Portfolio optimization algorithms for hierarchical risk parity.

This module implements various portfolio optimization algorithms:
- risk_parity: The main hierarchical risk parity algorithm
- one_over_n: A simple equal-weight allocation strategy
"""

from __future__ import annotations

from copy import deepcopy
from typing import Any, Generator

import pandas as pd

from .cluster import Cluster, Portfolio


def risk_parity(root: Cluster, cov: pd.DataFrame) -> Cluster:
    """
    Compute hierarchical risk parity weights for a cluster tree.

    This is the main algorithm for hierarchical risk parity. It recursively
    traverses the cluster tree and assigns weights to each node based on
    the risk parity principle.

    Args:
        root (Cluster): The root node of the cluster tree
        cov (pd.DataFrame): Covariance matrix of asset returns

    Returns:
        Cluster: The root node with portfolio weights assigned
    """
    if root.is_leaf:
        # a node is a leaf if has no further relatives downstream.
        asset = cov.keys().to_list()[root.value]
        root.portfolio[asset] = 1.0
        return root

    # drill down on the left
    root.left = risk_parity(root.left, cov)
    # drill down on the right
    root.right = risk_parity(root.right, cov)

    # combine left and right into a new cluster
    return _parity(root, cov=cov)


def _parity(cluster: Cluster, cov: pd.DataFrame) -> Cluster:
    """
    Compute risk parity weights for a parent cluster from its children.

    This function implements the core risk parity principle: allocating weights
    inversely proportional to risk, so that each sub-portfolio contributes
    equally to the total portfolio risk.

    Args:
        cluster (Cluster): The parent cluster with left and right children
        cov (pd.DataFrame): Covariance matrix of asset returns

    Returns:
        Cluster: The parent cluster with portfolio weights assigned
    """
    # Calculate variances of left and right sub-portfolios
    v_left = cluster.left.portfolio.variance(cov)
    v_right = cluster.right.portfolio.variance(cov)

    # Calculate weights inversely proportional to risk
    # such that v_left * alpha_left == v_right * alpha_right and alpha_left + alpha_right = 1
    alpha_left = v_right / (v_left + v_right)
    alpha_right = v_left / (v_left + v_right)

    # Combine assets from left and right clusters with their adjusted weights
    assets = {
        **(alpha_left * cluster.left.portfolio.weights).to_dict(),
        **(alpha_right * cluster.right.portfolio.weights).to_dict(),
    }

    # Assign the combined weights to the parent cluster's portfolio
    for asset, weight in assets.items():
        cluster.portfolio[asset] = weight

    return cluster


def one_over_n(dendrogram: Any) -> Generator[tuple[int, Portfolio]]:
    """
    Generate portfolios using the 1/N (equal weight) strategy at each tree level.

    This function implements a hierarchical 1/N strategy where weights are
    distributed equally among assets within each cluster at each level of the tree.
    The weight assigned to each cluster decreases by half at each level.

    Args:
        dendrogram: A dendrogram object containing the hierarchical clustering tree
                   and the list of assets

    Yields:
        tuple[int, Portfolio]: A tuple containing the level number and the portfolio
                              at that level
    """
    root = dendrogram.root
    assets = dendrogram.assets

    # Initial weight to distribute
    w = 1

    # Process each level of the tree
    for n, level in enumerate(root.levels):
        for node in level:
            # Distribute weight equally among all leaves in this node
            for leaf in node.leaves:
                root.portfolio[assets[leaf.value]] = w / node.leaf_count

        # Reduce weight for the next level
        w *= 0.5

        # Yield the current level number and a deep copy of the portfolio
        yield n, deepcopy(root.portfolio)



from pathlib import Path

import pandas as pd
import pytest

from pyhrp.cluster import Asset


@pytest.fixture(scope="session", name="resource_dir")
def resource_fixture() -> Path:
    """resource fixture"""
    return Path(__file__).parent / "resources"


@pytest.fixture(name="root_dir")
def root_fixture(resource_dir: Path) -> Path:
    return resource_dir.parent.parent.parent


@pytest.fixture(scope="session")
def prices(resource_dir: Path) -> pd.DataFrame:
    _prices = pd.read_csv(resource_dir / "stock_prices.csv", parse_dates=True, index_col="date").truncate(
        before="2017-01-01"
    )

    _prices.columns = [Asset(name=column) for column in _prices.columns]
    return _prices


@pytest.fixture(scope="session")
def returns(prices: pd.DataFrame) -> pd.DataFrame:
    return prices.pct_change().dropna(axis=0, how="all").fillna(0.0)



"""
Data structures for hierarchical risk parity portfolio optimization.

This module defines the core data structures used in the hierarchical risk parity algorithm:
- Asset: Represents a financial asset in a portfolio
- Portfolio: Manages a collection of assets and their weights
- Cluster: Represents a node in the hierarchical clustering tree
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

import numpy as np
import pandas as pd
from binarytree import Node


@dataclass(frozen=True)
class Asset:
    """
    Represents a financial asset in a portfolio.

    Attributes:
        mu (float, optional): Expected return of the asset
        name (str, optional): Name of the asset
    """

    mu: float = None
    name: str = None

    def __hash__(self) -> int:
        """
        Hash function for Asset objects.

        Returns:
            int: Hash value based on the asset name
        """
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        """
        Equality comparison for Asset objects.

        Args:
            other (Any): Object to compare with

        Returns:
            bool: True if other is an Asset with the same name
        """
        if not isinstance(other, Asset):
            return False
        return self.name == other.name

    def __lt__(self, other: Any) -> bool:
        """
        Less than comparison for Asset objects.

        Args:
            other (Any): Object to compare with

        Returns:
            bool: True if this asset's name is lexicographically less than other's name
        """
        if not isinstance(other, Asset):
            return False
        return self.name < other.name


@dataclass
class Portfolio:
    _weights: dict[Asset, float] = field(default_factory=dict)

    @property
    def assets(self) -> list[Asset]:
        """
        Get all assets in the portfolio.

        Returns:
            list[Asset]: List of assets in the portfolio
        """
        return list(self._weights.keys())

    def variance(self, cov: pd.DataFrame) -> float:
        """
        Calculate the variance of the portfolio.

        Args:
            cov (pd.DataFrame): Covariance matrix

        Returns:
            float: Portfolio variance
        """
        c = cov[self.assets].loc[self.assets].values
        w = self.weights[self.assets].values
        return np.linalg.multi_dot((w, c, w))

    def __getitem__(self, item: Asset) -> float:
        """
        Get the weight of an asset.

        Args:
            item (Asset): The asset to get the weight for

        Returns:
            float: The weight of the asset
        """
        return self._weights[item]

    def __setitem__(self, key: Asset, value: float) -> None:
        """
        Set the weight of an asset.

        Args:
            key (Asset): The asset to set the weight for
            value (float): The weight to set
        """
        self._weights[key] = value

    @property
    def weights(self) -> pd.Series:
        """
        Get all weights as a pandas Series.

        Returns:
            pd.Series: Series of weights indexed by assets
        """
        return pd.Series(self._weights, name="Weights").sort_index()

    def weights_by_name(self, names: list[str]) -> pd.Series:
        """
        Get weights as a pandas Series indexed by asset names.

        Args:
            names (list[str]): List of asset names to include

        Returns:
            pd.Series: Series of weights indexed by asset names
        """
        w = {asset.name: weight for asset, weight in self.weights.items()}
        return pd.Series({name: w[name] for name in names})

    def plot(self, names: list[str]):
        """
        Plot the portfolio weights.

        Args:
            names (list[str]): List of asset names to include in the plot

        Returns:
            matplotlib.axes.Axes: The plot axes
        """
        a = self.weights_by_name(names)

        ax = a.plot(kind="bar", color="skyblue")

        # Set x-axis labels and rotations
        ax.set_xticklabels(names, rotation=90, fontsize=8)
        return ax


class Cluster(Node):
    """
    Represents a cluster in the hierarchical clustering tree.

    Clusters are the nodes of the graphs we build.
    Each cluster is aware of the left and the right cluster
    it is connecting to. Each cluster also has an associated portfolio.

    Attributes:
        portfolio (Portfolio): The portfolio associated with this cluster
    """

    def __init__(self, value: int, left: Cluster | None = None, right: Cluster | None = None, **kwargs):
        """
        Initialize a new Cluster.

        Args:
            value (int): The identifier for this cluster
            left (Cluster, optional): The left child cluster
            right (Cluster, optional): The right child cluster
            **kwargs: Additional arguments to pass to the parent class
        """
        super().__init__(value=value, left=left, right=right, **kwargs)
        self.portfolio = Portfolio()

    @property
    def is_leaf(self) -> bool:
        """
        Check if this cluster is a leaf node (has no children).

        Returns:
            bool: True if this is a leaf node, False otherwise
        """
        return self.left is None and self.right is None

    @property
    def leaves(self) -> list[Cluster]:
        """
        Get all reachable leaf nodes in the correct order.

        Note that the leaves method of the Node class implemented in BinaryTree
        is not respecting the 'correct' order of the nodes.

        Returns:
            list[Cluster]: List of all leaf nodes reachable from this cluster
        """
        if self.is_leaf:
            return [self]
        else:
            return self.left.leaves + self.right.leaves
